from collections import defaultdict
from operator import itemgetter
from typing import Callable

import numpy as np

from .base import ScoredShot, Shot, Store, Vector


__all__ = ["MemoryStore"]


def cosine_distance(a: Vector, b: Vector) -> float:
    if not (norm_a := np.linalg.norm(a)) or not (norm_b := np.linalg.norm(b)):
        return 1.0
    return 1 - np.dot(a, b) / (norm_a * norm_b)


class MemoryStore(Store):
    # namespace => id => (shot, vector)
    _storage: dict[str, dict[str, tuple[Shot, Vector]]]
    distance: Callable[[Vector, Vector], float]

    def __init__(self, distance: Callable[[Vector, Vector], float] = cosine_distance):
        self._storage = defaultdict(dict)
        self.distance = distance

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        for shot, vector in zip(shots, vectors):
            self._storage[namespace][shot.id] = (shot, vector)

    def get(self, ids: list[str], namespace: str) -> list[Shot]:
        results: list[Shot] = []
        for id in ids:
            result = self._storage[namespace].get(id)
            if result:
                results.append(result[0])
        return results

    def remove(self, ids: list[str], namespace: str):
        for id in ids:
            del self._storage[namespace][id]

    def clear(self, namespace: str):
        self._storage[namespace].clear()

    def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        scored_shots = [
            ScoredShot(self.distance(vector, emb), shot)
            for (shot, emb) in self._storage[namespace].values()
        ]
        return sorted(scored_shots, key=itemgetter(1), reverse=True)[:limit]



from chromadb import Collection
from chromadb.api.async_client import AsyncCollection
from sorcery import dict_of

from few_shots.types import (
    dump_io_value,
    parse_io_value,
    ScoredShot,
    Shot,
    Vector,
)
from few_shots.utils.datetime import utcnow

from .base import AsyncStore, Store

__all__ = ["ChromaStore", "AsyncChromaStore"]


class ChromaStore(Store):
    collection: Collection

    def __init__(self, collection: Collection):
        self.collection = collection

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        self.collection.upsert(**ChromaHelper.upsert_shots(shots, vectors, namespace))

    def get(self, ids: list[str], _namespace: str) -> list[Shot]:
        return ChromaHelper.get_shots(self.collection.get(ids))

    def remove(self, ids: list[str], _namespace: str):
        self.collection.delete(ids=ids)

    def clear(self, namespace: str):
        self.collection.delete(where=dict_of(namespace))

    def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        return ChromaHelper.query_scored_shots(
            self.collection.query(
                query_embeddings=[vector],
                where=dict_of(namespace),
                n_results=limit,
            )
        )


class AsyncChromaStore(AsyncStore):
    collection: AsyncCollection

    def __init__(self, collection: AsyncCollection):
        self.collection = collection

    async def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        await self.collection.upsert(**ChromaHelper.upsert_shots(shots, vectors, namespace))

    async def get(self, ids: list[str], _namespace: str) -> list[Shot]:
        return ChromaHelper.get_shots(await self.collection.get(ids))

    async def remove(self, ids: list[str], _namespace: str):
        await self.collection.delete(ids=ids)

    async def clear(self, namespace: str):
        await self.collection.delete(where=dict_of(namespace))

    async def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        return ChromaHelper.query_scored_shots(
            await self.collection.query(
                query_embeddings=[vector],
                where=dict_of(namespace),
                n_results=limit,
            )
        )


class ChromaHelper:
    @staticmethod
    def upsert_shots(shots: list[Shot], vectors: list[Vector], namespace: str) -> dict:
        updated_at = utcnow()
        return dict(
            ids=[s.id for s in shots],
            embeddings=vectors,
            documents=[s.key for s in shots],
            metadatas=[
                dict_of(namespace, updated_at, outputs=dump_io_value(s.outputs)) for s in shots
            ],
        )

    @staticmethod
    def get_shots(results: dict) -> list[Shot]:
        return [
            Shot(parse_io_value(inputs), parse_io_value(metadata["outputs"]), id)
            for (inputs, metadata, id) in zip(
                results["documents"],
                results["metadatas"],
                results["ids"],
            )
        ]

    @staticmethod
    def query_scored_shots(results: dict) -> list[ScoredShot]:
        return [
            ScoredShot(
                distance,
                Shot(parse_io_value(inputs), parse_io_value(metadata["outputs"]), id),
            )
            for (distance, id, inputs, metadata) in zip(
                results["distances"][0],
                results["ids"][0],
                results["documents"][0],
                results["metadatas"][0],
            )
        ]



from abc import abstractmethod

from few_shots.types import Vector


class Embed:
    @abstractmethod
    def __call__(self, inputs: list[str]) -> list[Vector]: ...


class AsyncEmbed(Embed):
    @abstractmethod
    async def __call__(self, inputs: list[str]) -> list[Vector]: ...



import pytest

from few_shots.utils.asyncio import (
    asyncify_class,
    syncify_class,
    is_target,
    iscoroutinefunction,
)


def test_is_target():
    assert is_target("__call__") is True
    assert is_target("public_method") is True
    assert is_target("_private_method") is False


class SyncClass:
    def __call__(self, x):
        return x * 2

    def method(self, x):
        return x + 1

    def _private(self):
        return "private"


@asyncify_class
class AsyncClass(SyncClass):
    pass


@pytest.mark.asyncio
async def test_asyncify_class():
    obj = AsyncClass()

    result = await obj(2)
    assert result == 4, "__call__ should become async"

    result = await obj.method(1)
    assert result == 2, "public method should become async"

    assert not iscoroutinefunction(obj._private), "private method should remain sync"


class AsyncBaseClass:
    async def __call__(self, x):
        return x * 2

    async def method(self, x):
        return x + 1

    async def _private(self):
        return "private"


@syncify_class
class SyncFromAsyncClass(AsyncBaseClass):
    pass


def test_syncify_class():
    obj = SyncFromAsyncClass()

    assert obj(2) == 4, "__call__ should become sync"

    assert obj.method(1) == 2, "Public method should become sync"

    assert iscoroutinefunction(obj._private), "Private method should remain async"



from functools import partial
from typing import Any, Awaitable, Callable

from few_shots.types import Vector
from .base import AsyncEmbed, Embed


class OpenAIEmbed(Embed):
    def __init__(self, embedder: Callable[[str], Any], model: str, **kwargs):
        self.embedder = partial(embedder, model=model, **kwargs)

    def __call__(self, inputs: list[str]) -> list[Vector]:
        response = self.embedder(inputs)
        return [r["embedding"] for r in response["data"]]


class AsyncOpenAIEmbed(AsyncEmbed):
    def __init__(self, embedder: Callable[[str], Awaitable[Any]], model: str, **kwargs):
        self.embedder = partial(embedder, model=model, **kwargs)

    async def __call__(self, inputs: list[str]) -> list[Vector]:
        response = await self.embedder(inputs)
        return [r["embedding"] for r in response["data"]]



from sorcery import dict_of
from weaviate import WeaviateAsyncClient, WeaviateClient
from weaviate.classes.config import Property, DataType, VectorDistances, Configure
from weaviate.classes.data import DataObject
from weaviate.classes.query import Filter
from weaviate.collections.classes.internal import QueryReturnType
from weaviate.exceptions import WeaviateBaseError

from few_shots.types import (
    dump_io_value,
    parse_io_value,
    ScoredShot,
    Shot,
    Vector,
)
from few_shots.utils.datetime import utcnow

from .base import AsyncStore, Store

__all__ = ["WeaviateStore", "AsyncWeaviateStore", "VectorDistances"]


class WeaviateStore(Store):
    client: WeaviateClient
    collection_name: str
    distance_metric: VectorDistances

    def __init__(
        self,
        client: WeaviateClient,
        collection_name: str,
        distance_metric: VectorDistances = VectorDistances.COSINE,
    ):
        self.client = client
        self.collection_name = collection_name
        self.distance_metric = distance_metric

    def setup(self):
        try:
            self.collection = self.client.collections.create(
                **WeaviateHelper.collection_config(self.collection_name, self.distance_metric)
            )
        except WeaviateBaseError:
            self.collection = self.client.collections.get(self.collection_name)

    def teardown(self):
        self.client.collections.delete(self.collection_name)

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        self.collection.data.insert_many(WeaviateHelper.upsert_shots(shots, vectors, namespace))

    def get(self, ids: list[str], _namespace: str):
        return WeaviateHelper.fetch_shots(ids, self.collection.query.fetch_objects_by_ids(ids))

    def remove(self, ids: list[str], _namespace: str):
        self.collection.data.delete_many(Filter.by_id().contains_any(ids))

    def clear(self, namespace: str):
        self.collection.data.delete_many(Filter.by_property("namespace").equal(namespace))

    def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        return WeaviateHelper.query_scored_shots(
            self.collection.query.near_vector(
                vector,
                filters=Filter.by_property("namespace").equal(namespace),
                limit=limit,
            )
        )


class AsyncWeaviateStore(AsyncStore):
    client: WeaviateAsyncClient
    collection_name: str
    distance_metric: VectorDistances

    def __init__(
        self,
        client: WeaviateAsyncClient,
        collection_name: str,
        distance_metric: VectorDistances = VectorDistances.COSINE,
    ):
        self.client = client
        self.collection_name = collection_name
        self.distance_metric = distance_metric

    async def setup(self):
        try:
            self.collection = await self.client.collections.create(
                **WeaviateHelper.collection_config(self.collection_name, self.distance_metric)
            )
        except WeaviateBaseError:
            self.collection = await self.client.collections.get(self.collection_name)

    async def teardown(self):
        await self.client.collections.delete(self.collection_name)

    async def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        await self.collection.data.insert_many(
            WeaviateHelper.upsert_shots(shots, vectors, namespace)
        )

    async def get(self, ids: list[str], _namespace: str):
        return WeaviateHelper.fetch_shots(
            ids, await self.collection.query.fetch_objects_by_ids(ids)
        )

    async def remove(self, ids: list[str], _namespace: str):
        await self.collection.data.delete_many(Filter.by_id().contains_any(ids))

    async def clear(self, namespace: str):
        await self.collection.data.delete_many(Filter.by_property("namespace").equal(namespace))

    async def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        return WeaviateHelper.query_scored_shots(
            await self.collection.query.near_vector(
                vector,
                filters=Filter.by_property("namespace").equal(namespace),
                limit=limit,
            )
        )


class WeaviateHelper:
    @staticmethod
    def collection_config(collection_name: str, distance_metric: VectorDistances) -> dict:
        return dict(
            name=collection_name,
            properties=[
                Property(
                    name="namespace",
                    data_type=DataType.TEXT,
                    index_filterable=True,
                ),
                Property(name="inputs", data_type=DataType.TEXT),
                Property(name="outputs", data_type=DataType.TEXT),
                Property(name="updated_at", data_type=DataType.NUMBER),
            ],
            vector_index_config=Configure.VectorIndex.hnsw(distance_metric=distance_metric),
        )

    @staticmethod
    def upsert_shots(
        shots: list[Shot],
        vectors: list[Vector],
        namespace: str,
    ) -> list[DataObject]:
        updated_at = utcnow()
        return [
            DataObject(
                uuid=shot.id,
                vector=vector,
                properties=dict_of(
                    namespace,
                    updated_at,
                    inputs=dump_io_value(shot.inputs),
                    outputs=dump_io_value(shot.outputs),
                ),
            )
            for shot, vector in zip(shots, vectors)
        ]

    @staticmethod
    def fetch_shots(ids: list[str], response: QueryReturnType) -> list[Shot]:
        """
        Weaviate returns objects in a random order, so we need to order them by the ids we requested.
        """
        shots = {
            str(o.uuid): Shot(
                parse_io_value(o.properties["inputs"]),
                parse_io_value(o.properties["outputs"]),
                str(o.uuid),
            )
            for o in response.objects
        }
        return [shots[id] for id in ids if id in shots]

    @staticmethod
    def query_scored_shots(response: QueryReturnType) -> list[ScoredShot]:
        return [
            ScoredShot(
                o.metadata.distance,
                Shot(
                    parse_io_value(o.properties["inputs"]),
                    parse_io_value(o.properties["outputs"]),
                    str(o.uuid),
                ),
            )
            for o in response.objects
        ]



from hypothesis import given
from hypothesis.strategies import dictionaries, text

from few_shots.types import Shot, dump_io_value, id_io_value, is_io_value


@given(value=text())
def test_str_is_io_value(value: str):
    assert is_io_value(value)


@given(value=dictionaries(text(), text()))
def test_dict_is_io_value(value: dict[str, str]):
    assert is_io_value(value)


def test_data_key():
    data = {"b": 2, "a": 1}
    assert dump_io_value(data) == '{"a":1,"b":2}'


def test_data_hash(snapshot):
    data = {"test": "value"}
    assert id_io_value(data) == snapshot


@given(inputs=dictionaries(text(), text()))
def test_shot_custom_id(inputs: dict[str, str]):
    outputs = {"output": "result"}
    shot = Shot(inputs, outputs, "test_id")
    assert shot.inputs == inputs
    assert shot.outputs == outputs
    assert shot.key == dump_io_value(inputs)
    assert shot.id == "test_id"


@given(inputs=dictionaries(text(), text()))
def test_shot_auto_id(inputs: dict[str, str]):
    outputs = {"output": "result"}
    shot = Shot(inputs, outputs)
    assert shot.key == dump_io_value(inputs)
    assert shot.id == id_io_value(inputs)



from fastembed import TextEmbedding, SparseTextEmbedding

from few_shots.types import Vector

from .base import Embed


class FastEmbed(Embed):
    model: TextEmbedding | SparseTextEmbedding

    def __init__(self, model: TextEmbedding | SparseTextEmbedding):
        self.model = model

    def __call__(self, inputs: list[str]) -> list[Vector]:
        return [v.tolist() for v in self.model.embed(inputs)]



from dataclasses import dataclass
from functools import cached_property
from typing import NamedTuple, TypeVar
from uuid import uuid5, NAMESPACE_OID

import ujson


IO = TypeVar("IO", bound=dict | str)
Datum = TypeVar("Datum", bound=tuple[IO, IO] | tuple[IO, IO, str])
Vector = TypeVar("Vector", bound=list[float])


def is_io_value(value) -> bool:
    return isinstance(value, (dict, str))


def parse_io_value(value: str) -> IO:
    try:
        return ujson.loads(value)
    except ujson.JSONDecodeError:
        return value


def dump_io_value(data: IO) -> str:
    if isinstance(data, str):
        return data
    return ujson.dumps(data, sort_keys=True)


def id_io_value(data: IO) -> str:
    return str(uuid5(NAMESPACE_OID, dump_io_value(data)))


@dataclass
class Shot:
    inputs: IO
    outputs: IO
    id: str

    def __init__(self, inputs: IO, outputs: IO, id: str = ""):
        super().__init__()
        self.inputs = inputs
        self.outputs = outputs
        self.id = id or id_io_value(inputs)

    @cached_property
    def key(self) -> str:
        return dump_io_value(self.inputs)


ScoredShot = NamedTuple("ScoredShot", [("score", float), ("shot", Shot)])



from typing import Literal, TypeVar

import turbopuffer as tpuf

from few_shots.types import Shot, Vector, ScoredShot
from few_shots.utils.asyncio import asyncify_class
from few_shots.utils.datetime import utcnow

from .base import Store

__all__ = ["DistanceMetric", "TurboPufferStore", "AsyncTurboPufferStore"]

DistanceMetric = TypeVar("DistanceMetric", bound=Literal["cosine_distance", "euclidean_squared"])


class TurboPufferStore(Store):
    distance_metric: DistanceMetric

    def __init__(self, distance_metric: DistanceMetric = "cosine_distance"):
        self.distance_metric = distance_metric

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        updated_at = utcnow()
        tpuf.Namespace(namespace).upsert(
            ids=[shot.id for shot in shots],
            vectors=vectors,
            attributes={
                "inputs": [shot.input for shot in shots],
                "outputs": [shot.output for shot in shots],
                "updated_at": [updated_at for _ in shots],
            },
        )

    def get(self, ids: list[str], namespace: str) -> list[Shot]:
        raise NotImplementedError("TurboPuffer does not support get")

    def remove(self, ids: list[str], namespace: str):
        tpuf.Namespace(namespace).delete(ids)

    def clear(self, namespace: str):
        tpuf.Namespace(namespace).delete_all()

    def list(self, query_vector: Vector, namespace: str, limit: int):
        vector_results = tpuf.Namespace(namespace).query(
            vector=query_vector,
            top_k=limit,
            distance_metric=self.distance_metric,
            include_attributes=["inputs", "outputs"],
        )
        return [
            ScoredShot(
                row.dist,
                Shot(
                    id=row.id,
                    input=row.attributes["inputs"],
                    output=row.attributes["outputs"],
                ),
            )
            for row in vector_results
        ]


@asyncify_class
class AsyncTurboPufferStore(TurboPufferStore): ...



"""
Since some stores are stateful, make sure to clear them at the beginning of each test.
"""

import pytest
from pytest_lazy_fixtures import lf

from few_shots.store.base import Store, AsyncStore
from few_shots.types import Shot, Vector


providers = ["chroma", "pg", "qdrant", "weaviate"]  # TODO: Add Milvus & TurboPuffer


lazy_sync_stores = [lf(f"{p}_store") for p in providers]
lazy_async_stores = [lf(f"async_{p}_store") for p in providers]


@pytest.mark.parametrize("store", lazy_sync_stores)
def test_crud(
    store: Store,
    str_shots: list[Shot],
    mock_vectors: list[Vector],
    namespace: str,
):
    store.clear(namespace)

    store.add(str_shots, mock_vectors, namespace)
    store.add(str_shots, mock_vectors, namespace)  # Test idempotency

    query_vector = mock_vectors[0]
    (_d0, s0), (_d1, s1) = store.list(query_vector, namespace, limit=2)
    assert [s0, s1] == str_shots

    assert [s0, s1] == store.get([s0.id, s1.id], namespace)

    store.remove([s0.id], namespace)
    results = store.list(query_vector, namespace, limit=2)
    assert len(results) == 1
    assert results[0].shot == s1
    assert [] == store.get([s0.id], namespace)

    store.clear(namespace)
    results = store.list(query_vector, namespace, limit=2)
    assert len(results) == 0
    assert [] == store.get([s0.id], namespace)


@pytest.mark.parametrize("store", lazy_sync_stores)
def test_structured_io(
    store: Store,
    struct_shots: list[Shot],
    mock_vectors: list[Vector],
    namespace: str,
):
    store.clear(namespace)

    store.add(struct_shots, mock_vectors, namespace)

    query_vector = mock_vectors[0]
    (_d0, s0), (_d1, s1) = store.list(query_vector, namespace, limit=2)

    assert [s0, s1] == struct_shots
    assert [s0, s1] == store.get([s0.id, s1.id], namespace)


@pytest.mark.asyncio
@pytest.mark.parametrize("store", lazy_async_stores)
async def test_async_crud(
    store: AsyncStore,
    str_shots: list[Shot],
    mock_vectors: list[Vector],
    namespace: str,
):
    await store.clear(namespace)
    await store.add(str_shots, mock_vectors, namespace)
    await store.add(str_shots, mock_vectors, namespace)

    query_vector = mock_vectors[0]
    (_d0, s0), (_d1, s1) = await store.list(query_vector, namespace, limit=2)
    assert [s0, s1] == str_shots
    assert [s0, s1] == await store.get([s0.id, s1.id], namespace)

    await store.remove([s0.id], namespace)
    results = await store.list(query_vector, namespace, limit=2)
    assert len(results) == 1
    assert results[0].shot == s1
    assert [] == (await store.get([s0.id], namespace))

    await store.clear(namespace)
    results = await store.list(query_vector, namespace, limit=2)
    assert len(results) == 0
    assert [] == await store.get([s0.id], namespace)


@pytest.mark.asyncio
@pytest.mark.parametrize("store", lazy_async_stores)
async def test_async_structured_io(
    store: AsyncStore,
    struct_shots: list[Shot],
    mock_vectors: list[Vector],
    namespace: str,
):
    await store.clear(namespace)
    await store.add(struct_shots, mock_vectors, namespace)

    query_vector = mock_vectors[0]
    (_d0, s0), (_d1, s1) = await store.list(query_vector, namespace, limit=2)

    assert [s0, s1] == struct_shots
    assert [s0, s1] == await store.get([s0.id, s1.id], namespace)



from dataclasses import dataclass
from typing import overload

from few_shots.types import (
    Datum,
    dump_io_value,
    IO,
    id_io_value,
    is_io_value,
    ScoredShot,
    Shot,
)

from .embed.base import AsyncEmbed
from .store.base import AsyncStore


@dataclass
class AsyncFewShots:
    """
    Client for storing and retrieving few-shot examples.

    Combines an embedding model with a vector store to enable semantic search over examples.

    `add` is used to store examples, `remove` to delete them, `clear` to remove all examples in a namespace, and `list` to find similar examples to an input.
    """

    embed: AsyncEmbed
    store: AsyncStore

    @overload
    async def add(
        self,
        inputs: IO,
        outputs: IO,
        *,
        id: str = "",
        namespace: str = "default",
    ) -> str:
        """Add an example to the store.

        Args:
            inputs: Input to store
            outputs: Output to store
            id: ID for the example
            namespace: Namespace to store the example in
        """

    @overload
    async def add(
        self,
        data: list[Datum],
        *,
        namespace: str = "default",
    ) -> list[str]:
        """Add multiple examples to the store.

        Args:
            data: List of (input, output) or (input, output, id) tuples
            namespace: Namespace to store the examples in
        """

    async def add(
        self,
        maybe_inputs: IO | list[Datum],
        maybe_outputs: IO | None = None,
        *,
        id: str = "",
        namespace: str = "default",
    ) -> str | list[str]:
        is_io_args = is_io_value(maybe_inputs) and is_io_value(maybe_outputs)
        data: list[Datum] = [(maybe_inputs, maybe_outputs, id)] if is_io_args else maybe_inputs
        shots = [Shot(*datum) for datum in data]
        vectors = await self.embed([shot.key for shot in shots])
        await self.store.add(shots, vectors, namespace)

        ids = [shot.id for shot in shots]
        return ids[0] if is_io_args else ids

    @overload
    async def get(self, inputs: IO, *, namespace: str = "default") -> Shot | None:
        """Get a single shot from the store, useful for retrieving known good outputs.

        Args:
            inputs: Input to look up
            namespace: Namespace to get the example from

        Returns:
            Shot: The example with the given input
        """

    @overload
    async def get(self, inputs: list[IO], *, namespace: str = "default") -> list[Shot]:
        """Get multiple shots from the store, useful for retrieving known good outputs.

        Args:
            inputs: List of inputs to look up
            namespace: Namespace to get the examples from

        Returns:
            list[Shot]: The examples with the given inputs
        """

    async def get(
        self,
        inputs: IO | list[IO],
        *,
        namespace: str = "default",
    ) -> Shot | list[Shot]:
        is_single = not isinstance(inputs, list)
        if is_single:
            inputs = [inputs]

        ids = [id_io_value(i) for i in inputs]
        shots = await self.store.get(ids, namespace)

        if is_single:
            return shots[0] if shots else None

        return shots

    @overload
    async def remove(self, ids: list[str], *, namespace: str = "default"):
        """Remove one or more examples from the store.

        Args:
            ids: IDs of examples to remove
            namespace: Namespace to remove examples from
        """

    @overload
    async def remove(
        self,
        inputs: IO,
        outputs: IO | None = None,
        *,
        id: str = "",
        namespace: str = "default",
    ):
        """Remove one example from the store.

        Args:
            inputs: Input of example to remove
            outputs: Output of example to remove
            id: ID of example to remove
            namespace: Namespace to remove the example from
        """

    @overload
    async def remove(
        self,
        data: list[Datum],
        namespace: str = "default",
    ):
        """Remove multiple examples from the store.

        Args:
            data: List of (input, output) or (input, output, id) tuples
            namespace: Namespace to remove the examples from
        """

    async def remove(
        self,
        maybe_inputs: IO | list[Datum],
        maybe_outputs: IO | None = None,
        *,
        id: str = "",
        namespace: str = "default",
    ):
        is_io_args = is_io_value(maybe_inputs) and is_io_value(maybe_outputs)
        data: list[Datum] = [(maybe_inputs, maybe_outputs, id)] if is_io_args else maybe_inputs
        ids = data if isinstance(data[0], str) else [Shot(*datum).id for datum in data]
        await self.store.remove(ids, namespace)

    async def clear(self, namespace: str = "default"):
        """Remove all examples from a namespace.

        Args:
            namespace: Namespace to clear
        """
        await self.store.clear(namespace)

    async def list(
        self,
        inputs: IO,
        *,
        namespace: str = "default",
        limit: int = 5,
    ) -> list[ScoredShot]:
        """Find similar examples to an input.

        Args:
            inputs: Input to find similar examples for
            namespace: Namespace to search in
            limit: Maximum number of examples to return

        Returns:
            List of (example, score) tuples, sorted by distance ascending
        """
        [vector] = await self.embed([dump_io_value(inputs)])
        return await self.store.list(vector, namespace, limit)



from sentence_transformers import SentenceTransformer

from few_shots.types import Vector
from .base import Embed


class TransformersEmbed(Embed):
    model: SentenceTransformer

    def __init__(self, model: SentenceTransformer):
        self.model = model

    def __call__(self, inputs: list[str]) -> list[Vector]:
        return self.model.encode(inputs).tolist()



import pytest

from few_shots.client import FewShots
from few_shots.store.memory import MemoryStore
from few_shots.types import Shot


@pytest.fixture
def client():
    def embed(inputs: list[str]):
        return [[1] * 384] * len(inputs)

    return FewShots(embed=embed, store=MemoryStore())


def test_functional_flow(client: FewShots):
    inputs = {"a": 1}
    outputs = {"b": 2}
    shot = Shot(inputs, outputs)

    id = client.add(inputs, outputs)
    assert id == shot.id

    results = [r.shot for r in client.list(inputs)]
    assert results == [shot]
    assert shot == client.get(inputs)

    client.remove(inputs, outputs)
    assert [] == client.list(inputs)
    assert client.get(inputs) is None

    client.add(inputs, outputs)
    client.remove(inputs, outputs)
    assert [] == client.list(inputs)
    assert client.get(inputs) is None


def test_dispatch(client: FewShots):
    inputs = {"a": 1}
    outputs = {"b": 2}

    id = client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    client.remove([id])
    assert [] == client.list(inputs)

    [id] = client.add([(inputs, outputs)])
    client.remove([id])
    assert [] == client.list(inputs)

    client.add([(inputs, outputs)])
    client.remove([(inputs, outputs)])
    assert [] == client.list(inputs)

    [id] = client.add([(inputs, outputs, "id")])
    assert id == "id"

    client.remove([(inputs, outputs, "id")])
    assert [] == client.list(inputs)

    client.add([(inputs, outputs, "id")])
    client.remove(["id"])
    assert [] == client.list(inputs)


def test_string_flow(client: FewShots):
    inputs = "User question..."
    outputs = "AI answer..."
    shot = Shot(inputs, outputs)

    id = client.add(inputs, outputs)
    assert id == shot.id

    results = [r.shot for r in client.list(inputs, limit=1)]
    assert results == [shot]
    assert client.get(inputs) == shot

    results = [r.shot for r in client.list(inputs, limit=5)]
    assert results == [shot]

    client.remove(inputs, outputs)
    assert [] == [r.shot for r in client.list(inputs, limit=1)]

    client.add(inputs, outputs)
    client.clear()
    assert [] == [r.shot for r in client.list(inputs)]
    assert client.get(inputs) is None



from typing import Literal, TypeVar
from pymilvus import (
    CollectionSchema,
    DataType,
    FieldSchema,
    MilvusClient,
)

from few_shots.types import (
    dump_io_value,
    parse_io_value,
    ScoredShot,
    Shot,
    Vector,
)
from few_shots.utils.asyncio import asyncify_class

from .base import Store


MetricType = TypeVar("MetricType", bound=Literal["L2", "IP", "COSINE", "HAMMING", "JACCARD"])


class MilvusStore(Store):
    client: MilvusClient
    collection_name: str

    def __init__(self, client: MilvusClient, collection_name: str):
        self.client = client
        self.collection_name = collection_name

    def setup(self, size: int, metric_type: MetricType = "COSINE"):
        if self.client.has_collection(self.collection_name):
            return

        fields = [
            FieldSchema("id", DataType.VARCHAR, max_length=128, is_primary=True),
            FieldSchema("namespace", DataType.VARCHAR, max_length=512),
            FieldSchema("vector", DataType.FLOAT_VECTOR, dim=size),
            FieldSchema("payload", DataType.JSON),
            FieldSchema("updated_at", DataType.FLOAT),
        ]

        self.client.create_collection(
            collection_name=self.collection_name,
            schema=CollectionSchema(fields),
            index=self.client.prepare_index_params(
                field_name="vector",
                metric_type=metric_type,
                index_type="AUTOINDEX",
                index_name="vectors_index",
            ),
        )

    def teardown(self):
        self.client.drop_collection(self.collection_name)

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        self.client.upsert(
            collection_name=self.collection_name,
            data=[
                {
                    "id": shot.id,
                    "namespace": namespace,
                    "vectors": vector,
                    "payload": {
                        "inputs": dump_io_value(shot.inputs),
                        "outputs": dump_io_value(shot.outputs),
                    },
                }
                for shot, vector in zip(shots, vectors)
            ],
        )

    def get(self, ids: list[str], namespace: str) -> list[Shot]:
        response = self.client.query(
            collection_name=self.collection_name,
            filter=f"namespace == '{namespace}'",
            output_fields=["id", "payload"],
            ids=ids,
        )[0]

        results: list[Shot] = []
        for datum in response:
            metadata = datum.get("entity", {}).get("payload")
            inputs = parse_io_value(metadata["inputs"])
            outputs = parse_io_value(metadata["outputs"])
            results.append(Shot(inputs, outputs, datum["id"]))

        return results

    def remove(self, ids: list[str], namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            ids=ids,
            filter=f"namespace == '{namespace}'",
        )

    def clear(self, namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            filter=f"namespace == '{namespace}'",
        )

    def list(
        self,
        vector: Vector,
        namespace: str,
        limit: int,
    ) -> list[ScoredShot]:
        response = self.client.search(
            collection_name=self.collection_name,
            data=[vector],
            filter=f"namespace == '{namespace}'",
            limit=limit,
        )[0]

        results: list[ScoredShot] = []
        for datum in response:
            metadata = datum.get("entity", {}).get("payload")
            inputs = parse_io_value(metadata["inputs"])
            outputs = parse_io_value(metadata["outputs"])
            results.append(ScoredShot(datum["distance"], Shot(inputs, outputs, datum["id"])))

        return results


@asyncify_class
class AsyncMilvusStore(MilvusStore): ...



from typing import Literal, TypeVar
from uuid import UUID
from psycopg import AsyncConnection, Connection
from psycopg.rows import TupleRow
from psycopg.types.json import Jsonb
from pgvector.psycopg import register_vector, register_vector_async

from few_shots.types import ScoredShot, Shot, Vector

from .base import Store


__all__ = ["PGStore", "AsyncPGStore"]


DistanceType = TypeVar("DistanceType", bound=Literal["cosine", "l2"])


class PGStore(Store):
    _sql: "SQLHelper"
    connection: Connection

    def __init__(
        self,
        connection: Connection[TupleRow],
        tablename: str,
        schema: str = "public",
        distance: DistanceType = "cosine",
    ):
        self._sql = SQLHelper(tablename, schema, distance)
        self.connection = connection

    def setup(self, dimensions: int, m: int = 64, ef_construction: int = 128):
        """
        Sets up the database table and index for vector similarity search.
        Idempotent, will not re-create the table if it already exists.
        Will use DiskANN as an index if available, HNSW otherwise.

        Args:
            dimensions: The number of dimensions in the vectors to be stored
            m: Number of connections per element in HNSW index (pgvector only)
            ef_construction: Size of the dynamic candidate list for constructing HNSW index (pgvector only)

        Raises:
            ValueError: If pgvector extension is not installed in the database
        """
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.pgvector_create())
            cursor.execute(self._sql.table_create(dimensions))

            cursor.execute(self._sql.diskann_check())
            if cursor.fetchone():
                cursor.execute(self._sql.diskann_index())
            else:
                cursor.execute(self._sql.pgvector_index(m, ef_construction))

        register_vector(self.connection)

    def teardown(self):
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.table_drop())

    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        with self.connection.cursor() as cursor:
            cursor.executemany(
                self._sql.upsert(),
                self._sql.upsert_shots(shots, vectors, namespace),
            )

    def get(self, ids: list[str], _namespace: str) -> list[Shot]:
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.select(), (ids,))
            return self._sql.select_shots(cursor.fetchall())

    def remove(self, ids: list[str], _namespace: str):
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.remove(), (ids,))

    def clear(self, namespace: str):
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.clear(), (namespace,))

    def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]:
        with self.connection.cursor() as cursor:
            cursor.execute(self._sql.query(), (vector, namespace, limit))
            return self._sql.query_scored_shots(cursor.fetchall())


class AsyncPGStore(Store):
    _sql: "SQLHelper"
    connection: AsyncConnection

    def __init__(
        self,
        connection: AsyncConnection[TupleRow],
        tablename: str,
        schema: str = "public",
        distance: DistanceType = "cosine",
    ):
        self._sql = SQLHelper(tablename, schema, distance)
        self.connection = connection

    async def setup(self, dimensions: int, m: int = 64, ef_construction: int = 128):
        """
        Sets up the database table and index for vector similarity search.
        Idempotent, will not re-create the table if it already exists.
        Will use DiskANN as an index if available, HNSW otherwise.
        """
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.pgvector_create())
            await cursor.execute(self._sql.table_create(dimensions))

            await cursor.execute(self._sql.diskann_check())
            if await cursor.fetchone():
                await cursor.execute(self._sql.diskann_index())
            else:
                await cursor.execute(self._sql.pgvector_index(m, ef_construction))

        await register_vector_async(self.connection)

    async def teardown(self):
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.table_drop())

    async def add(self, shots: list[Shot], vectors: list[Vector], namespace: str):
        async with self.connection.cursor() as cursor:
            await cursor.executemany(
                self._sql.upsert(),
                self._sql.upsert_shots(shots, vectors, namespace),
            )

    async def get(self, ids: list[str], _namespace: str) -> list[Shot]:
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.select(), (ids,))
            return self._sql.select_shots(await cursor.fetchall())

    async def remove(self, ids: list[str], _namespace: str):
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.remove(), (ids,))

    async def clear(self, namespace: str):
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.clear(), (namespace,))

    async def list(
        self,
        vector: Vector,
        namespace: str,
        limit: int,
    ) -> list[ScoredShot]:
        async with self.connection.cursor() as cursor:
            await cursor.execute(self._sql.query(), (vector, namespace, limit))
            return self._sql.query_scored_shots(await cursor.fetchall())


class SQLHelper:
    tablename: str
    schema: str
    distance: DistanceType

    def __init__(
        self,
        tablename: str,
        schema: str = "public",
        distance: DistanceType = "cosine",
    ):
        self.tablename = tablename
        self.schema = schema
        self.distance = distance

    @staticmethod
    def utcnow():
        return "timezone('utc', now())"

    def table_create(self, vector_dimensions: int):
        return f"""\
        CREATE TABLE IF NOT EXISTS {self.schema}.{self.tablename} (
            id UUID PRIMARY KEY,
            namespace VARCHAR(256) NOT NULL,
            payload JSONB NOT NULL,
            vector VECTOR({vector_dimensions}) NOT NULL,
            updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT {self.utcnow()}
        );
        """

    def table_drop(self):
        return f"DROP TABLE IF EXISTS {self.schema}.{self.tablename};"

    @staticmethod
    def pgvector_check():
        return "SELECT * FROM pg_extension WHERE extname = 'vector'"

    @staticmethod
    def pgvector_create():
        return "CREATE EXTENSION IF NOT EXISTS vector"

    def pgvector_index(self, m: int, ef_construction: int):
        return f"""\
        CREATE INDEX IF NOT EXISTS index_{self.tablename}_vector
        ON {self.schema}.{self.tablename}
        USING hnsw (vector vector_{self.distance}_ops)
        WITH (m = {m}, ef_construction = {ef_construction});
        """

    @staticmethod
    def diskann_check():
        return "SELECT * FROM pg_extension WHERE extname = 'vectorscale';"

    @staticmethod
    def diskann_create():
        return "CREATE EXTENSION IF NOT EXISTS vectorscale;"

    def diskann_index(self):
        return f"""\
        CREATE INDEX IF NOT EXISTS index_{self.tablename}_vector
        ON {self.schema}.{self.tablename}
        USING DISKANN (vector);
        """

    def upsert(self):
        return f"""\
        INSERT INTO {self.schema}.{self.tablename} (id, namespace, payload, vector, updated_at)
        VALUES (%s, %s, %s, %s, {self.utcnow()})
        ON CONFLICT (id) DO UPDATE SET
            namespace = EXCLUDED.namespace,
            payload = EXCLUDED.payload,
            vector = EXCLUDED.vector,
            updated_at = {self.utcnow()};
        """

    def upsert_shots(
        self,
        shots: list[Shot],
        vectors: list[Vector],
        namespace: str,
    ) -> list[tuple[str, str, Jsonb, Vector]]:
        return [
            (
                shot.id,
                namespace,
                Jsonb({"inputs": shot.inputs, "outputs": shot.outputs}),
                vector,
            )
            for shot, vector in zip(shots, vectors)
        ]

    def select(self):
        return f"""\
        SELECT {self.tablename}.id,
               {self.tablename}.payload
        FROM {self.schema}.{self.tablename}
        WHERE {self.tablename}.id = ANY(%s);
        """

    def select_shots(self, tuples: list[tuple[UUID, dict]]) -> list[Shot]:
        return [Shot(payload["inputs"], payload["outputs"], str(id)) for (id, payload) in tuples]

    def query(self):
        return f"""\
        SELECT {self.tablename}.id,
               {self.tablename}.payload,
               {self.tablename}.vector <-> %s::vector AS distance
        FROM {self.schema}.{self.tablename}
        WHERE {self.tablename}.namespace = %s
        ORDER BY distance ASC
        LIMIT %s;
        """

    def query_scored_shots(
        self,
        tuples: list[tuple[UUID, dict, float]],
    ) -> list[ScoredShot]:
        return [
            ScoredShot(
                distance,
                Shot(payload["inputs"], payload["outputs"], str(id)),
            )
            for (id, payload, distance) in tuples
        ]

    def remove(self):
        return f"DELETE FROM {self.schema}.{self.tablename} WHERE id = ANY(%s)"

    def clear(self):
        return f"DELETE FROM {self.schema}.{self.tablename} WHERE namespace = %s"



import pytest

from few_shots.async_client import AsyncFewShots
from few_shots.store.memory import MemoryStore
from few_shots.types import Shot
from few_shots.utils.asyncio import asyncify_class


@asyncify_class
class AsyncMemoryStore(MemoryStore):
    """
    Async version of MemoryStore for testing only.
    """


@pytest.fixture(scope="function")
def client():
    async def embed(inputs: list[str]):
        return [[1] * 384] * len(inputs)

    return AsyncFewShots(embed=embed, store=AsyncMemoryStore())


@pytest.mark.asyncio
async def test_functional_flow(client: AsyncFewShots):
    inputs = {"a": 1}
    outputs = {"b": 2}
    shot = Shot(inputs, outputs)

    id = await client.add(inputs, outputs)
    assert id == shot.id

    results = [r.shot for r in await client.list(inputs)]
    assert results == [shot]
    assert shot == await client.get(inputs)

    await client.remove(inputs, outputs)
    assert [] == await client.list(inputs)

    await client.add(inputs, outputs)
    await client.clear()
    assert [] == await client.list(inputs)
    assert (await client.get(inputs)) is None


@pytest.mark.asyncio
async def test_dispatch(client: AsyncFewShots):
    inputs = {"a": 1}
    outputs = {"b": 2}
    shot = Shot(inputs, outputs)

    id = await client.add(inputs, outputs)
    assert id == shot.id
    assert shot == await client.get(inputs)
    assert [shot] == await client.get([inputs])

    await client.remove([id])
    assert [] == await client.list(inputs)

    [id] = await client.add([(inputs, outputs)])
    await client.remove([id])
    assert [] == await client.list(inputs)

    await client.add([(inputs, outputs)])
    await client.remove([(inputs, outputs)])
    assert [] == await client.list(inputs)

    [id] = await client.add([(inputs, outputs, "id")])
    assert id == "id"

    await client.remove([(inputs, outputs, "id")])
    assert [] == await client.list(inputs)

    await client.add([(inputs, outputs, "id")])
    await client.remove(["id"])
    assert [] == await client.list(inputs)


@pytest.mark.asyncio
async def test_string_flow(client: AsyncFewShots):
    inputs = "User question..."
    outputs = "AI answer..."
    shot = Shot(inputs, outputs)

    id = await client.add(inputs, outputs)
    assert id == shot.id
    assert shot == await client.get(inputs)

    results = [s.shot for s in await client.list(inputs, limit=5)]
    assert [shot] == results

    await client.remove(inputs, outputs)
    assert [] == await client.list(inputs, limit=1)

    await client.add(inputs, outputs)
    await client.clear()
    assert [] == await client.list(inputs)
    assert (await client.get(inputs)) is None



from asyncio import iscoroutinefunction
from typing import TypeVar

from asyncer import asyncify, syncify

C = TypeVar("C")


def is_target(method_name: str) -> bool:
    return method_name == "__call__" or not method_name.startswith("_")


def syncify_class(cls: C) -> C:
    for base in [cls] + list(cls.__bases__):
        for name, method in base.__dict__.items():
            if iscoroutinefunction(method) and is_target(name):
                setattr(cls, name, syncify(method, raise_sync_error=False))
    return cls


def asyncify_class(cls: C) -> C:
    for base in [cls] + list(cls.__bases__):
        for name, method in base.__dict__.items():
            if not iscoroutinefunction(method) and is_target(name):
                setattr(cls, name, asyncify(method))
    return cls



from itertools import chain
from typing import overload

from few_shots.types import ScoredShot, Shot, dump_io_value


@overload
def shots_to_messages(scored_shots: list[ScoredShot]) -> list[dict]: ...


@overload
def shots_to_messages(shots: list[Shot]) -> list[dict]: ...


def shots_to_messages(shots: list[ScoredShot] | list[Shot]) -> list[dict]:
    if not shots:
        return []

    sample = shots[0]
    if isinstance(sample, ScoredShot):
        shots = [s.shot for s in shots]

    return list(
        chain.from_iterable(
            [
                {"role": "user", "content": dump_io_value(shot.inputs)},
                {"role": "assistant", "content": dump_io_value(shot.outputs)},
            ]
            for shot in shots
        )
    )



import pytest

from chromadb import HttpClient, AsyncHttpClient
from psycopg import AsyncConnection, Connection
from pymilvus import MilvusClient
from qdrant_client.local.async_qdrant_local import AsyncQdrantLocal
from qdrant_client.local.qdrant_local import QdrantLocal
import weaviate

from few_shots.store.chroma import ChromaStore, AsyncChromaStore
from few_shots.store.milvus import AsyncMilvusStore, MilvusStore
from few_shots.store.pg import AsyncPGStore, PGStore
from few_shots.store.qdrant import AsyncQdrantStore, QdrantStore, Distance
from few_shots.store.weaviate import AsyncWeaviateStore, WeaviateStore
from few_shots.types import Shot, Vector


# Shared fixtures
@pytest.fixture
def str_shots():
    return [Shot("input1", "output1"), Shot("input2", "output2")]


@pytest.fixture
def struct_shots():
    return [
        Shot({"key": "input1"}, {"key": "output1"}),
        Shot({"key": "input2"}, {"key": "output2"}),
    ]


@pytest.fixture
def mock_vectors() -> list[Vector]:
    return [[1.0, 2.0], [3.0, 4.0]]


@pytest.fixture
def namespace() -> str:
    return "test"


# Chroma fixtures
@pytest.fixture
def chroma_store():
    client = HttpClient()
    collection = client.create_collection("test", get_or_create=True)
    return ChromaStore(collection)


@pytest.fixture
async def async_chroma_store():
    client = await AsyncHttpClient()
    collection = await client.create_collection("test", get_or_create=True)
    return AsyncChromaStore(collection)


# PostgreSQL fixtures
@pytest.fixture
def pg_conn():
    with Connection.connect(
        "postgresql://postgres:postgres@localhost:5432/postgres"
    ) as c:
        yield c


@pytest.fixture
def pg_store(pg_conn: Connection):
    s = PGStore(connection=pg_conn, tablename="few_shots")
    s.setup(dimensions=2)
    yield s
    s.teardown()


# PostgreSQL async fixtures
@pytest.fixture
async def async_pg_conn():
    async with await AsyncConnection.connect(
        "postgresql://postgres:postgres@localhost:5432/postgres"
    ) as c:
        yield c


@pytest.fixture
async def async_pg_store(async_pg_conn: AsyncConnection):
    s = AsyncPGStore(connection=async_pg_conn, tablename="few_shots")
    await s.setup(dimensions=2)
    yield s
    await s.teardown()


# Qdrant fixtures
@pytest.fixture
def qdrant_store():
    s = QdrantStore(client=QdrantLocal(":memory:"), collection_name="test")
    s.setup(size=2, distance=Distance.COSINE)
    yield s
    s.teardown()


# Qdrant async fixtures
@pytest.fixture
async def async_qdrant_store():
    s = AsyncQdrantStore(client=AsyncQdrantLocal(":memory:"), collection_name="test")
    await s.setup(size=2, distance=Distance.COSINE)
    yield s
    await s.teardown()


# Weaviate fixtures
@pytest.fixture
def weaviate_client():
    with weaviate.connect_to_local() as c:
        yield c


@pytest.fixture
def weaviate_store(weaviate_client):
    s = WeaviateStore(weaviate_client, "test")
    s.setup()
    yield s
    s.teardown()


# Weaviate async fixtures
@pytest.fixture
async def async_weaviate_client():
    async with weaviate.use_async_with_local() as c:
        yield c


@pytest.fixture
async def async_weaviate_store(async_weaviate_client):
    s = AsyncWeaviateStore(async_weaviate_client, "test")
    await s.setup()
    yield s
    await s.teardown()


@pytest.fixture
def milvus_store():
    s = MilvusStore(MilvusClient(), "test")
    s.setup()
    yield s
    s.teardown()


@pytest.fixture
async def async_milvus_store():
    s = AsyncMilvusStore(MilvusClient(), "test")
    await s.setup()
    yield s
    await s.teardown()



from abc import abstractmethod

from few_shots.types import Vector, Shot, ScoredShot


class Store:
    @abstractmethod
    def add(self, shots: list[Shot], vectors: list[Vector], namespace: str): ...

    @abstractmethod
    def get(self, ids: list[str], namespace: str) -> list[Shot]: ...

    @abstractmethod
    def remove(self, ids: list[str], namespace: str): ...

    @abstractmethod
    def clear(self, namespace: str): ...

    @abstractmethod
    def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]: ...


class AsyncStore(Store):
    @abstractmethod
    async def add(self, shots: list[Shot], vectors: list[Vector], namespace: str): ...

    @abstractmethod
    async def get(self, ids: list[str], namespace: str) -> list[Shot]: ...

    @abstractmethod
    async def remove(self, ids: list[str], namespace: str): ...

    @abstractmethod
    async def clear(self, namespace: str): ...

    @abstractmethod
    async def list(self, vector: Vector, namespace: str, limit: int) -> list[ScoredShot]: ...



from few_shots.types import Shot
from few_shots.utils.format import shots_to_messages


def test_shots_to_messages_empty():
    assert shots_to_messages([]) == []


def test_shots_to_messages_single_shot():
    shot = Shot(inputs="test input", outputs="test output")
    expected = [
        {"role": "user", "content": "test input"},
        {"role": "assistant", "content": "test output"},
    ]
    assert shots_to_messages([shot]) == expected


def test_shots_to_messages_multiple_shots():
    shots = [
        Shot(inputs="input 1", outputs="output 1"),
        Shot(inputs="input 2", outputs="output 2"),
    ]
    expected = [
        {"role": "user", "content": "input 1"},
        {"role": "assistant", "content": "output 1"},
        {"role": "user", "content": "input 2"},
        {"role": "assistant", "content": "output 2"},
    ]
    assert shots_to_messages(shots) == expected


def test_shots_to_messages_dict_inputs():
    shot = Shot(
        inputs={"context": "some context", "prompt": "test input"},
        outputs="test output",
    )
    expected = [
        {
            "content": '{"context":"some context","prompt":"test input"}',
            "role": "user",
        },
        {"content": "test output", "role": "assistant"},
    ]
    assert shots_to_messages([shot]) == expected


def test_shots_to_messages_dict_outputs():
    shot = Shot(
        inputs="test input",
        outputs={"metadata": {"confidence": 0.9}, "response": "test output"},
    )
    expected = [
        {"content": "test input", "role": "user"},
        {
            "content": '{"metadata":{"confidence":0.9},"response":"test output"}',
            "role": "assistant",
        },
    ]
    assert shots_to_messages([shot]) == expected


def test_shots_to_messages_dict_both():
    shot = Shot(inputs={"prompt": "test input"}, outputs={"response": "test output"})
    expected = [
        {"content": '{"prompt":"test input"}', "role": "user"},
        {"content": '{"response":"test output"}', "role": "assistant"},
    ]
    assert shots_to_messages([shot]) == expected



from typing import List

from qdrant_client import QdrantClient, AsyncQdrantClient
from qdrant_client.models import (
    Distance,
    FieldCondition,
    Filter,
    HasIdCondition,
    HnswConfigDiff,
    KeywordIndexParams,
    MatchValue,
    PointStruct,
    Record,
    ScoredPoint,
    VectorParams,
)
from sorcery import dict_of

from few_shots.types import (
    dump_io_value,
    parse_io_value,
    ScoredShot,
    Shot,
    Vector,
)
from few_shots.utils.datetime import utcnow

from .base import AsyncStore, Store

__all__ = ["QdrantStore", "AsyncQdrantStore", "Distance"]


class QdrantStore(Store):
    client: QdrantClient
    collection_name: str

    def __init__(self, client: QdrantClient, collection_name: str):
        self.client = client
        self.collection_name = collection_name

    def setup(self, size: int, distance: Distance, payload_m: int = 16):
        self.client.create_collection(
            **QdrantHelper.create_collection(self.collection_name, size, distance, payload_m)
        )
        self.client.create_payload_index(
            **QdrantHelper.create_payload_index(self.collection_name),
        )

    def teardown(self):
        self.client.delete_collection(self.collection_name)

    def add(self, shots: List[Shot], vectors: List[Vector], namespace: str):
        self.client.upsert(
            collection_name=self.collection_name,
            points=QdrantHelper.upsert_points(shots, vectors, namespace),
        )

    def get(self, ids: List[str], _namespace: str) -> List[Shot]:
        return QdrantHelper.retrieve_shots(
            self.client.retrieve(collection_name=self.collection_name, ids=ids)
        )

    def remove(self, ids: List[str], namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=QdrantHelper.selector(namespace, ids),
        )

    def clear(self, namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=QdrantHelper.selector(namespace),
        )

    def list(self, vector: Vector, namespace: str, limit: int) -> List[ScoredShot]:
        results = self.client.search(
            collection_name=namespace,
            query_vector=vector,
            query_filter=QdrantHelper.selector(namespace),
            limit=limit,
        )
        return QdrantHelper.search_scored_shots(results)


class AsyncQdrantStore(AsyncStore):
    client: AsyncQdrantClient
    collection_name: str

    def __init__(self, client: AsyncQdrantClient, collection_name: str):
        self.client = client
        self.collection_name = collection_name

    async def setup(self, size: int, distance: Distance, payload_m: int = 16):
        await self.client.create_collection(
            **QdrantHelper.create_collection(self.collection_name, size, distance, payload_m)
        )
        await self.client.create_payload_index(
            **QdrantHelper.create_payload_index(self.collection_name),
        )

    async def teardown(self):
        await self.client.delete_collection(self.collection_name)

    async def add(self, shots: List[Shot], vectors: List[Vector], namespace: str):
        await self.client.upsert(
            collection_name=namespace,
            points=QdrantHelper.upsert_points(shots, vectors, namespace),
        )

    async def get(self, ids: List[str], _namespace: str) -> List[Shot]:
        return QdrantHelper.retrieve_shots(
            await self.client.retrieve(collection_name=self.collection_name, ids=ids)
        )

    async def remove(self, ids: List[str], namespace: str):
        await self.client.delete(
            collection_name=namespace,
            points_selector=QdrantHelper.selector(namespace, ids),
        )

    async def clear(self, namespace: str):
        await self.client.delete(
            collection_name=namespace,
            points_selector=QdrantHelper.selector(namespace),
        )

    async def list(self, vector: Vector, namespace: str, limit: int) -> List[ScoredShot]:
        results = await self.client.search(
            collection_name=namespace,
            query_vector=vector,
            query_filter=QdrantHelper.selector(namespace),
            limit=limit,
        )
        return QdrantHelper.search_scored_shots(results)


class QdrantHelper:
    @staticmethod
    def create_collection(collection_name: str, size: int, distance: Distance, payload_m: int = 16):
        """
        Use these as kwargs for `.create_collection` for optimal performance.
        """
        return dict(
            collection_name=collection_name,
            vectors_config=VectorParams(size=size, distance=distance),
            hnsw_config=HnswConfigDiff(payload_m=payload_m, m=0),
        )

    @staticmethod
    def create_payload_index(collection_name: str):
        """
        Use these as kwargs for `.create_payload_index` for optimal performance.
        """
        return dict(
            collection_name=collection_name,
            field_name="namespace",
            field_schema=KeywordIndexParams(type="keyword", is_tenant=True),
        )

    @staticmethod
    def selector(namespace: str, ids: list[str] | None = None) -> Filter:
        cond = (
            FieldCondition(key="namespace", match=MatchValue(value=namespace))
            if not ids
            else HasIdCondition(has_id=ids)
        )
        return Filter(must=[cond])

    @staticmethod
    def upsert_points(
        shots: List[Shot],
        vectors: List[Vector],
        namespace: str,
    ) -> List[PointStruct]:
        updated_at = utcnow()
        return [
            PointStruct(
                id=shot.id,
                vector=vector,
                payload=dict_of(
                    namespace,
                    updated_at,
                    inputs=dump_io_value(shot.inputs),
                    outputs=dump_io_value(shot.outputs),
                ),
            )
            for shot, vector in zip(shots, vectors)
        ]

    @staticmethod
    def retrieve_shots(results: List[Record]) -> list[Shot]:
        return [
            Shot(
                inputs=parse_io_value(result.payload["inputs"]),
                outputs=parse_io_value(result.payload["outputs"]),
                id=str(result.id),
            )
            for result in results
        ]

    @staticmethod
    def search_scored_shots(results: List[ScoredPoint]) -> List[ScoredShot]:
        return [
            ScoredShot(
                result.score,
                Shot(
                    id=result.id,
                    inputs=parse_io_value(result.payload["inputs"]),
                    outputs=parse_io_value(result.payload["outputs"]),
                ),
            )
            for result in results
        ]



from .async_client import AsyncFewShots
from .client import FewShots
from .types import Shot, Vector, ScoredShot, IO, Datum
from .utils.format import shots_to_messages

__all__ = [
    "AsyncFewShots",
    "FewShots",
    "Datum",
    "Vector",
    "IO",
    "Shot",
    "shots_to_messages",
    "ScoredShot",
]
